pub trait Inlining: Sized {
    fn iread_n<A, E>(
        self,
        n: usize,
        ok: impl FnOnce(&[u8]) -> A,
        err: impl FnOnce(&[u8]) -> E,
    ) -> Result<(A, Self), E>;

    fn iread_all<A>(self, ok: impl FnOnce(&[u8]) -> A) -> A;

    fn itell(&self) -> usize;
}

pub trait InliningExt<E>: Inlining {
    fn iread_n_const<const N: usize>(
        self,
        err: impl FnOnce(&[u8]) -> E,
    ) -> Result<([u8; N], Self), E> {
        self.iread_n(
            N,
            |slice| slice.try_into().expect("iread_n wrong length."),
            err,
        )
    }
}

impl<D: Inlining, E> InliningExt<E> for D {}
